@@ -1,11 +1,11 @@
-use std::iter::repeat_with;
-
-use crate::analyzer::analyze;
-
 use super::Optimizer;
+use crate::analyzer::analyze;
+use crate::compress::optimize::is_left_access_to_arguments;
+use std::iter::repeat_with;
 use swc_atoms::js_word;
 use swc_common::DUMMY_SP;
 use swc_ecma_ast::*;
+use swc_ecma_utils::ident::IdentLike;
 use swc_ecma_utils::private_ident;
 use swc_ecma_visit::noop_visit_mut_type;
 use swc_ecma_visit::VisitMut;
@@ -53,7 +53,7 @@ impl Optimizer<'_> {
             return;
         }
 
-        if f.params.iter().any(|param| match param.pat {
+        if f.params.iter().any(|param| match &param.pat {
             Pat::Ident(BindingIdent {
                 id:
                     Ident {
@@ -62,7 +62,12 @@ impl Optimizer<'_> {
                     },
                 ..
             }) => true,
-            Pat::Ident(..) => false,
+            Pat::Ident(i) => self
+                .data
+                .as_ref()
+                .and_then(|v| v.vars.get(&i.id.to_id()))
+                .map(|v| v.declared_count >= 2)
+                .unwrap_or(false),
             _ => true,
         }) {
             return;
@@ -82,6 +87,7 @@ impl Optimizer<'_> {
             params: &mut f.params,
             changed: false,
             keep_fargs: self.options.keep_fargs,
+            prevent: false,
         };
 
         // We visit body two time, to use simpler logic in `inject_params_if_required`
@@ -96,6 +102,7 @@ struct ArgReplacer<'a> {
     params: &'a mut Vec<Param>,
     changed: bool,
     keep_fargs: bool,
+    prevent: bool,
 }
 
 impl ArgReplacer<'_> {
@@ -126,7 +133,19 @@ impl ArgReplacer<'_> {
 impl VisitMut for ArgReplacer<'_> {
     noop_visit_mut_type!();
 
+    fn visit_mut_assign_expr(&mut self, n: &mut AssignExpr) {
+        n.visit_mut_children_with(self);
+
+        if is_left_access_to_arguments(&n.left) {
+            self.prevent = true;
+        }
+    }
+
     fn visit_mut_expr(&mut self, n: &mut Expr) {
+        if self.prevent {
+            return;
+        }
+
         n.visit_mut_children_with(self);
 
         match n {
@@ -197,6 +216,10 @@ impl VisitMut for ArgReplacer<'_> {
     }
 
     fn visit_mut_member_expr(&mut self, n: &mut MemberExpr) {
+        if self.prevent {
+            return;
+        }
+
         n.obj.visit_mut_with(self);
 
         if n.computed {
